from bot_common.messages.timeout_model import TimeoutObj, timeout_headers as headers
from bot_common.utils.db_utils import DbConfig, create_table_if_not_exists, db_connect, update_table
from bot_common.utils.utils import catch_exception, get_time_now, clean_string
from bot_common.utils.logging_conf import logger
from bot_common.lib_conf import table_name_timeout

int_fields = ['timeout_sec']
clean_str_fields = ['company_contact', 'user_contact']


# TimeoutMessageManager class is only used by the message.py classes

class TimeoutMessageManager:
    def __init__(self, db, cursor, new_timeout: TimeoutObj = None):
        self.db, self.cursor = db, cursor
        self.tab_name = table_name_timeout
        self.new_timeout = new_timeout
        self.timeout_sec_idx = list(headers.keys()).index('timeout_sec') + 1
        create_table_if_not_exists(self.db, self.cursor, self.tab_name, headers)

    @catch_exception
    def get(self, required_elapsed_seconds_default):
        get_timeout_query = f"SELECT * FROM {self.tab_name} WHERE company_contact = '{self.new_timeout.company_contact}' AND user_contact = '{self.new_timeout.user_contact}';"
        self.cursor.execute(get_timeout_query)
        new_timeout_ls = self.cursor.fetchall()
        if new_timeout_ls:
            return new_timeout_ls[0][self.timeout_sec_idx]
        return required_elapsed_seconds_default

    @catch_exception
    def delete(self):
        delete_timeout_query = f"DELETE FROM {self.tab_name} WHERE company_contact = '{self.new_timeout.company_contact}' AND user_contact = '{self.new_timeout.user_contact}';"
        self.cursor.execute(delete_timeout_query)
        self.db.commit()

    @catch_exception
    def delete_older_mins(self, delete_timeout_before_tmp):
        delete_expired_query = f"DELETE FROM {self.tab_name} WHERE timestamp < '{delete_timeout_before_tmp}';"
        self.cursor.execute(delete_expired_query)
        self.db.commit()


@catch_exception
def set_timeout(new_timeout: TimeoutObj, db_config: DbConfig):
    db, cursor = db_connect(db_config)
    tab_name = table_name_timeout
    create_table_if_not_exists(db, cursor, tab_name, headers)

    data = new_timeout.__dict__.copy()
    data['timestamp'] = get_time_now()
    for field in int_fields:
        data[field] = int(data[field])
    for field in clean_str_fields:
        data[field] = clean_string(data[field])

    update_table(db, cursor, data, tab_name)
    logger.info(f'set new_timeout {data} success')
    db.close()



from bot_common.exceptions.exception_model import ExceptionObj, exception_headers as headers
from bot_common.utils.utils import catch_exception, get_time_now, clean_string, validate_time, datetime_to_str
from bot_common.utils.db_utils import DbConfig, create_table_if_not_exists, db_connect, insert_into_table
from bot_common.utils.logging_conf import logger
from bot_common.lib_conf import table_name_exceptions
from typing import List
import json

int_fields = []
clean_str_fields = ['error_message']


class ExceptionSession(ExceptionObj):
    def extract_data(self):
        data = self.__dict__.copy()
        data['timestamp'] = self.timestamp if validate_time(self.timestamp) else get_time_now()
        for field in int_fields:
            data[field] = int(data[field])
        for field in clean_str_fields:
            data[field] = clean_string(data[field])
        return data

    @catch_exception
    def write(self, db_config: DbConfig):
        tab_name = table_name_exceptions
        session_db, session_cursor = db_connect(db_config)
        create_table_if_not_exists(session_db, session_cursor, tab_name, headers)
        data_dict = self.extract_data()
        insert_into_table(session_db, session_cursor, data_dict, tab_name)
        logger.info(f'Writing exception success')
        session_db.close()


class ExceptionManager:
    @catch_exception
    def __init__(self, tmp: str, db_config: DbConfig):
        self.tmp = tmp
        self.tab_name = table_name_exceptions
        self.session_db, self.session_cursor = db_connect(db_config)
        create_table_if_not_exists(self.session_db, self.session_cursor, self.tab_name, headers)

    @catch_exception
    def delete(self):
        delete_session_query = f"DELETE FROM {self.tab_name} WHERE timestamp <= '{self.tmp}';"
        self.session_cursor.execute(delete_session_query)
        self.session_db.commit()
        logger.info(f'delete exceptions success')
        self.session_db.close()

    @catch_exception
    def get(self) -> List[ExceptionObj]:
        get_session_query = f"SELECT * FROM {self.tab_name} WHERE timestamp <= '{self.tmp}' ORDER BY timestamp DESC;"
        self.session_cursor.execute(get_session_query)
        myresult = self.session_cursor.fetchall()
        self.session_db.close()
        out_ls = []
        for res in myresult:
            # res[1:] because we need to remove the pk auto_increment element
            out_dict = dict(zip(headers, res[1:]))
            out_dict['timestamp'] = datetime_to_str(out_dict['timestamp'])
            out_ls.append(ExceptionObj.parse_obj(out_dict))
        return out_ls



from datetime import datetime, timezone
import re


def catch_exception(func):
    def wrapper(*args, **kwargs):
        try:
            out = func(*args, **kwargs)
            return out
        except Exception as e:
            f_name = func.__name__
            exc = f'{f_name} --> {str(e)}'
            raise Exception(exc)
    return wrapper


@catch_exception
def clean_string(txt):
    out = txt.replace(r"\n", " ").replace("'", " ")
    out = re.sub(r"\s+", " ", out)
    return out.strip()


@catch_exception
def get_time_now():
    return datetime.now(timezone.utc).astimezone().strftime("%Y-%m-%d %H:%M:%S.%f")


@catch_exception
def str_to_datetime(tmp_str):
    tmp_str1 = tmp_str.split('.')[0]
    tmp = datetime.strptime(tmp_str1, "%Y-%m-%d %H:%M:%S")
    return tmp


@catch_exception
def datetime_to_str(dt):
    out_str = dt.strftime("%Y-%m-%d %H:%M:%S.%f")
    return out_str


@catch_exception
def validate_time(date_str):
    try:
        datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S.%f")
        return True
    except ValueError:
        return False


@catch_exception
def remove_special_characters(txt):
    out = txt.replace(r"\n", " ")
    out = out.translate({ord(c): " " for c in r"""!@#$%^&*()'[]{};:,./<>?\|`"~-=_+"""})
    out = re.sub(r"\s+", " ", out)
    return out.strip()



from pydantic import BaseModel


class SendMailRequest(BaseModel):
    target_address_ls: list = []
    mail_title: str = ''
    mail_body: str = ''


class SenderMailConfig(BaseModel):
    sender_address: str
    sender_password: str
    provider_address: str = 'smtp.gmail.com'
    provider_port: int = 587



from pathlib import Path
from pydantic import BaseModel


class ConfigPaths(BaseModel):
    config_file_path: Path
    flow_path: Path
    responses_path: Path


# '<U_IDX>' flags all the fields that cannot be simultaneously duplicated,
# '<IDX>' flags all the fields that are used in the queries
flow_headers = {
                'in_ctx': 'VARCHAR(250) NOT NULL <U_IDX> <IDX>',
                'intent': 'VARCHAR(250) <U_IDX>',
                'entity': 'VARCHAR(250) <U_IDX>',
                'action': 'TEXT',
                'action_out': 'VARCHAR(250) <U_IDX>',
                'out_ctx': 'TEXT NOT NULL',
                'response_id': 'TEXT NOT NULL',
                'response_id_fallback': 'TEXT',
                'hints': 'TEXT',
                'timeout_sec': 'INT',
                'call_reason': 'TEXT'
                }

responses_headers = {'id': 'VARCHAR(250) NOT NULL <U_IDX> <IDX>',
                     'body1': 'TEXT',
                     'body2': 'TEXT',
                     'body3': 'TEXT',
                     'body4': 'TEXT',
                     'body5': 'TEXT'
                     }

config_headers = {
                  'company': 'VARCHAR(250) NOT NULL <U_IDX> <IDX>',
                  'is_active': 'TINYINT NOT NULL <IDX>',
                  'is_active_switch_tmp': 'VARCHAR(50)',
                  'company_port': 'INT NOT NULL',
                  'company_contact': 'VARCHAR(250) NOT NULL <IDX>',
                  'used_conversation_platform': 'TEXT',
                  'redirect_contacts': 'TEXT',
                  'nlu_port': 'INT NOT NULL',
                  'sessions_expiration_mins': 'INT',
                  'unexisting_session_start_ctx': 'TEXT',
                  'preferred_nlu_extractor': 'TEXT',
                  'default_timeout_sec': 'INT',
                  'fallback_counter_max': 'INT',
                  'technical_issue_response_id': 'TEXT',
                  'session_expired_response_id_by_start_ctx': 'TEXT',
                  'handover_incomprehension_response_id': 'TEXT',
                  'end_conversation_formalities_intents': 'TEXT',
                  'preprocessing_messages_contexts': 'TEXT',
                  'preprocessing_messages_bodies': 'TEXT',
                  'hints_mapping': 'TEXT',
                  'solicit_after_mins': 'TEXT',
                  'solicit_action_ls': 'TEXT',
                  'success_other_logs': 'TEXT',
                  'other': 'TEXT'
                  }





from bot_common.utils.db_utils import db_connect
from bot_common.utils.utils import catch_exception, clean_string, get_time_now, str_to_datetime, datetime_to_str, remove_special_characters
from bot_common.utils.logging_conf import logger



from bot_common.companies.company_model import config_headers as headers
from bot_common.utils.utils import catch_exception, get_time_now
from bot_common.utils.db_utils import DbConfig, db_connect, create_table_if_not_exists
from bot_common.utils.logging_conf import logger
from bot_common.lib_conf import table_name_company_configurations, table_prefix_conversational_flow, table_prefix_responses
from pydantic import BaseModel
import json
import pandas as pd


@catch_exception
def dict_parser(d: dict):
    parsed_dict = dict()
    for key, val in d.items():
        try:
            val_p = json.loads(val) if isinstance(val, str) else val
        except json.JSONDecodeError:
            val_p = val
        parsed_dict[key] = val_p
    return {k: v for k, v in parsed_dict.items() if v is not None}


class GenericObj(BaseModel):
    def __init__(self, d: dict):
        self.__dict__.update(dict_parser(d))


class Company:
    def __init__(self, field: str, value, db_config: DbConfig):
        self.field = field
        self.value = value
        self.tab_name = table_name_company_configurations
        self.db, self.cursor = db_connect(db_config)
        create_table_if_not_exists(self.db, self.cursor, self.tab_name, headers)

    @catch_exception
    def get_config(self, opened=True):
        where_active_field = 'AND is_active = 1' if opened else ''
        get_config_query = f"SELECT * FROM {self.tab_name} WHERE {self.field} LIKE '%{self.value}%' {where_active_field};"
        self.cursor.execute(get_config_query)
        conf_vals = self.cursor.fetchall()
        if not conf_vals:
            self.db.close()
            raise Exception(f'get company configuration - {self.field} {self.value} not found')
        elif len(conf_vals) > 1:
            self.db.close()
            raise Exception(f'get company configuration - {self.field} {self.value} ambiguous or duplicated')

        # get config structure
        get_structure_query = f"SELECT column_name FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = '{self.tab_name}' ORDER BY ordinal_position;"
        self.cursor.execute(get_structure_query)
        structure = self.cursor.fetchall()
        self.db.close()
        # flatten the structure list
        structure = list(sum(structure, ()))
        # remove duplicates from structure_result while keeping the order
        structure = list(dict.fromkeys(structure))
        config_dict = dict(zip(structure[1:], conf_vals[0][1:]))
        return GenericObj(config_dict)

    @catch_exception
    def delete_config(self):
        get_config_query = f"SELECT company FROM {self.tab_name} WHERE {self.field} LIKE '%{self.value}%';"
        self.cursor.execute(get_config_query)
        conf_vals = self.cursor.fetchall()
        if not conf_vals:
            self.db.close()
            raise Exception(f'delete company configuration - {self.field} {self.value} not found')
        elif len(conf_vals) > 1:
            self.db.close()
            raise Exception(f'delete company configuration - {self.field} {self.value} ambiguous or duplicated')

        company = conf_vals[0][0]
        delete_config_query = f"DELETE FROM {self.tab_name} WHERE {self.field} LIKE '%{self.value}%';"
        delete_flow_query = f"DROP TABLE {table_prefix_conversational_flow}{company}"
        delete_responses_query = f"DROP TABLE {table_prefix_responses}{company}"
        self.cursor.execute(delete_config_query)
        self.db.commit()
        self.cursor.execute(delete_flow_query)
        self.db.commit()
        self.cursor.execute(delete_responses_query)
        self.db.commit()
        logger.info(f'>> deleted configuration for company {company}')
        self.db.close()


class CompanyManager:
    def __init__(self, db_config: DbConfig):
        self.tab_name = table_name_company_configurations
        self.db, self.cursor = db_connect(db_config)
        create_table_if_not_exists(self.db, self.cursor, self.tab_name, headers)

    @catch_exception
    def get_companies(self, opened: bool = True):
        where_active_field = 'WHERE is_active = 1' if opened else ''
        get_companies_query = f"SELECT company FROM {self.tab_name} {where_active_field};"
        self.cursor.execute(get_companies_query)
        companies = self.cursor.fetchall()
        self.db.close()
        return [el[0] for el in companies]

    @catch_exception
    def disable_company(self, company: str):
        get_company_query = f"SELECT company FROM {self.tab_name} WHERE company = '{company}';"
        self.cursor.execute(get_company_query)
        if not self.cursor.fetchall():
            self.db.close()
            raise Exception(f'disable company - {company} not found')

        disable_config_query = f"UPDATE {self.tab_name} SET is_active = 0, is_active_switch_tmp = '{get_time_now()}' WHERE company = '{company}';"
        self.cursor.execute(disable_config_query)
        self.db.commit()
        logger.info(f'>> disabled {company} company')
        self.db.close()

    @catch_exception
    def activate_company(self, company: str):
        get_company_query = f"SELECT company FROM {self.tab_name} WHERE company = '{company}';"
        self.cursor.execute(get_company_query)
        if not self.cursor.fetchall():
            self.db.close()
            raise Exception(f'activate company - {company} not found')

        activate_config_query = f"UPDATE {self.tab_name} SET is_active = 1, is_active_switch_tmp = '{get_time_now()}' WHERE company = '{company}';"
        self.cursor.execute(activate_config_query)
        self.db.commit()
        logger.info(f'>> activated {company} company')
        self.db.close()


class CompanyTables:
    def __init__(self, db_config: DbConfig, company: str):
        self.company = company
        self.tab_name = ''
        self.db, self.cursor = db_connect(db_config)

    @catch_exception
    def get_flow_filtered_by_input_ctx(self, in_ctx: str):
        self.tab_name = table_prefix_conversational_flow + self.company
        get_table_query = f"SELECT * FROM {self.tab_name} WHERE in_ctx = '{in_ctx}';"
        df = pd.read_sql(get_table_query, self.db)
        self.db.close()
        df = df.drop(columns=['pk'])
        if len(df) == 0:
            raise Exception(f'Invalid input context {in_ctx}')
        return df

    @catch_exception
    def get_responses_csv(self):
        self.tab_name = table_prefix_responses + self.company
        get_table_query = f"SELECT * FROM {self.tab_name};"
        df = pd.read_sql(get_table_query, self.db)
        self.db.close()
        return df.drop(columns=['pk'])



from pydantic import BaseModel


class SessionObject(BaseModel):
    id_session: str
    is_closed: int = 0
    company: str
    company_contact: str
    start_conversation_timestamp: str
    user_key: str = ''
    start_context: str = ''
    current_context: str = ''
    state: str = ''
    timestamp: str = ''
    timeout_sec: int = 0
    bot_message_contains_buttons: bool = False
    unclosed_success: bool = False
    platform_exception: bool = False
    solicited_times: int = 0
    message_id: str = ''
    bot_message: str = ''
    other_logs: dict = {}
    extracted_data: dict = {}
    cache: dict = {}
    entities: dict = {}


# '<U_IDX>' flags all the fields that cannot be simultaneously duplicated,
# '<IDX>' flags all the fields that are used in the queries
session_headers = {'id_session': 'VARCHAR(250) NOT NULL <U_IDX>',
                   'is_closed': 'TINYINT NOT NULL <IDX>',
                   'company': 'TEXT NOT NULL',
                   'company_contact': 'TEXT NOT NULL',
                   'user_key': 'TEXT',
                   'start_context': 'TEXT',
                   'current_context': 'TEXT',
                   'state': 'TEXT',
                   'start_conversation_timestamp': 'VARCHAR(50)',
                   'timestamp': 'VARCHAR(50)',
                   'timeout_sec': 'TINYINT',
                   'bot_message_contains_buttons': 'TINYINT',
                   'unclosed_success': 'TINYINT',
                   'platform_exception': 'TINYINT',
                   'solicited_times': 'TINYINT',
                   'message_id': 'TEXT',
                   'bot_message': 'TEXT',
                   'other_logs': 'TEXT',
                   'extracted_data': 'TEXT',
                   'cache': 'TEXT',
                   'entities': 'TEXT'
                   }



from bot_common.utils.utils import catch_exception, get_time_now, clean_string, str_to_datetime, datetime_to_str
from bot_common.utils.db_utils import DbConfig, create_table_if_not_exists, db_connect, insert_into_table
from bot_common.logs.log_model import LogObject, log_headers as headers
from bot_common.utils.logging_conf import logger
from bot_common.lib_conf import table_prefix_logs, table_prefix_logs_end_conversation
from datetime import timedelta
import json

# - In "other_logs" we store some indicators about the performed
#   conversational flow path.
# - In "extracted_data" we store the main features extracted
#   during the conversation.

str_special_char_fields = []
float_fields = ['intent_confidence']
dump_fields = ['other_logs', 'extracted_data']
clean_str_fields = ['bot_message', 'current_user_utterance', 'log_transcript']
int_fields = ['conv_duration_sec', 'fallback', 'handover', 'handover_incomprehension',
              'conv_step_num', 'closed', 'expired', 'redirect', 'closed_formality',
              'hangup', 'unclosed_success', 'solicit', 'platform_exception']


class Log(LogObject):
    @catch_exception
    def extract_data(self):
        data = self.__dict__.copy()
        data['timestamp'] = get_time_now()
        for field in dump_fields:
            data[field] = clean_string(json.dumps(data[field])).replace('\\', '\\\\')
        for field in str_special_char_fields:
            data[field] = data[field].replace("\\", "\\\\").replace("'", "\\'")
        for field in int_fields:
            data[field] = int(data[field])
        for field in clean_str_fields:
            data[field] = clean_string(data[field])
        for field in float_fields:
            data[field] = round(float(data[field]), 2)
        return data

    @catch_exception
    def write(self, db_config: DbConfig, end_conversation=False):
        prefix = table_prefix_logs_end_conversation if end_conversation else table_prefix_logs
        tab_name = prefix + self.company
        logs_db, logs_cursor = db_connect(db_config)
        create_table_if_not_exists(logs_db, logs_cursor, tab_name, headers)
        data_dict = self.extract_data()
        try:
            insert_into_table(logs_db, logs_cursor, data_dict, tab_name)
            logger.info(f'Writing logs success')
        except Exception as err:
            try:
                time_new = datetime_to_str(str_to_datetime(get_time_now()) + timedelta(seconds=1))
                data_dict['timestamp'] = time_new
                insert_into_table(logs_db, logs_cursor, data_dict, tab_name)
                logger.info(f'Writing logs success (at second try) - {err}')
            except Exception as e:
                raise Exception(f'Exception in writing logs: {e}')
        finally:
            logs_db.close()



from bot_common.utils.db_utils import DbConfig, create_table_if_not_exists, db_connect, update_table, insert_into_table
from bot_common.utils.utils import catch_exception, get_time_now, clean_string, str_to_datetime
from bot_common.messages.message_model import MessageDbObj, NewMessageDbObj, DmMessageLightObj, message_headers as headers
from bot_common.messages.timeout_model import TimeoutObj
from bot_common.messages.timeout import TimeoutMessageManager
from bot_common.utils.logging_conf import logger
from bot_common.lib_conf import void_response, table_name_messages
from datetime import timedelta
from typing import List
import itertools

int_fields = []
clean_str_fields = ['company_contact', 'user_contact', 'user_message']


class NewMessages:
    def __init__(self, db_config: DbConfig):
        self.db, self.cursor = db_connect(db_config)
        self.tab_name = table_name_messages
        self.current_tmp = str_to_datetime(get_time_now())
        self.headers_keys = list(headers.keys())
        # +1 because we have the pk auto_increment element as first element
        self.company_idx = self.headers_keys.index('company_contact') + 1
        self.user_contact_idx = self.headers_keys.index('user_contact') + 1
        self.timestamp_idx = self.headers_keys.index('timestamp') + 1
        self.required_elapsed_seconds_default = 0
        self.single_user_parsed_messages = []
        self.single_user_messages = []
        self.single_message = []
        self.company_user_key = ()
        self.new_messages_dict = {}
        create_table_if_not_exists(self.db, self.cursor, self.tab_name, headers)

    @catch_exception
    def set_processed_and_parse(self):
        message_dict = dict(zip(self.headers_keys, self.single_message))
        message_dict['processed'] = 1
        message_obj = MessageDbObj.parse_obj(message_dict)
        self.single_user_parsed_messages.append(message_obj)
        # set processed=1 for the extracted messages
        update_table(self.db, self.cursor, message_dict, self.tab_name)

    @catch_exception
    def check_single_user_messages(self):
        user_last_received_tmp = str_to_datetime(self.single_user_messages[0][self.timestamp_idx])
        # time_elapsed_sec is the time elapsed from the given user's last message
        time_elapsed_sec = (self.current_tmp - user_last_received_tmp).total_seconds()
        # if it is passed a given amount of time since the user's last message,
        # we add the user messages between the processable ones, otherwise we wait for it
        to = TimeoutObj(company_contact=self.company_user_key[0], user_contact=self.company_user_key[1])
        required_elapsed_seconds = TimeoutMessageManager(self.db, self.cursor, to).get(self.required_elapsed_seconds_default)

        logger.info(f'required_elapsed_seconds: {required_elapsed_seconds}')
        if time_elapsed_sec > required_elapsed_seconds:
            TimeoutMessageManager(self.db, self.cursor, to).delete()
            self.single_user_parsed_messages = []
            for message in self.single_user_messages:
                # mess[1:] because we need to remove the pk auto_increment element
                self.single_message = message[1:]
                self.set_processed_and_parse()
            self.new_messages_dict[self.company_user_key] = self.single_user_parsed_messages

    @catch_exception
    def get(self, required_elapsed_seconds_default: int):
        self.required_elapsed_seconds_default = required_elapsed_seconds_default
        # query for the unprocessed messages:
        # ORDER BY company, user_contact --> in order to allow the grouping by user_contact for different companies
        # ORDER BY timestamp DESC (per each user_contact) --> in order to have the last received message as first record
        get_new_messages_query = f"SELECT * FROM {self.tab_name} WHERE processed = '0' ORDER BY company_contact, user_contact, timestamp DESC;"
        self.cursor.execute(get_new_messages_query)
        new_messages_ls = self.cursor.fetchall()
        # group by different user_contacts
        grouper = itertools.groupby(new_messages_ls, key=lambda x: (x[self.company_idx], x[self.user_contact_idx]))
        # cu_key = (company_contact, user_contact) tuple
        for cu_key, user_messages in grouper:
            logger.info(f'get_new_messages - (company_contact, user_contact): {cu_key}')
            self.company_user_key = cu_key
            self.single_user_messages = list(user_messages)
            self.check_single_user_messages()
        self.db.close()
        return self.new_messages_dict

    @catch_exception
    def set(self, new_message: NewMessageDbObj):
        data = new_message.__dict__.copy()
        data['timestamp'] = get_time_now()
        data['processed'] = 0
        del data['token']
        for field in int_fields:
            data[field] = int(data[field])
        for field in clean_str_fields:
            data[field] = clean_string(data[field])

        void_message_content = not bool(data.get('user_message')) and not bool(data.get('start_context'))
        data['user_message'] = list(void_response.values())[0] if void_message_content else data.get('user_message')
        insert_into_table(self.db, self.cursor, data, self.tab_name)
        logger.info(f'set_new_message {data} success')
        self.db.close()


# --------------- DELETE MESSAGES CLASS

class DeleteMessages:
    def __init__(self, db_config: DbConfig):
        self.db, self.cursor = db_connect(db_config)
        self.tab_name = table_name_messages
        self.company_contact = ''
        self.user_contact = ''
        self.current_tmp = str_to_datetime(get_time_now())
        create_table_if_not_exists(self.db, self.cursor, self.tab_name, headers)

    @catch_exception
    def expired(self, expired_ls: List[DmMessageLightObj]):
        for exp in expired_ls:
            self.company_contact = exp.company_contact
            self.user_contact = exp.session_id
            delete_expired_query = f"DELETE FROM {self.tab_name} WHERE company_contact = '{self.company_contact}' AND user_contact = '{self.user_contact}';"
            self.cursor.execute(delete_expired_query)
            self.db.commit()
            to = TimeoutObj(company_contact=self.company_contact, user_contact=self.user_contact)
            TimeoutMessageManager(self.db, self.cursor, to).delete()
            logger.info(f'deleted expired messages/timeouts ({self.company_contact}, {self.user_contact}) success')
        self.db.close()

    @catch_exception
    def older_mins(self, delete_older_messages_mins):
        timedelta_mins = timedelta(minutes=delete_older_messages_mins)
        delete_messages_before_tmp = self.current_tmp - timedelta_mins
        delete_expired_query = f"DELETE FROM {self.tab_name} WHERE timestamp < '{delete_messages_before_tmp}';"
        self.cursor.execute(delete_expired_query)
        self.db.commit()
        TimeoutMessageManager(self.db, self.cursor).delete_older_mins(delete_messages_before_tmp)
        logger.info('>>> DELETED OLDER MESSAGES/TIMEOUTS')
        self.db.close()



from pydantic import BaseModel


class TimeoutObj(BaseModel):
    company_contact: str
    user_contact: str
    timestamp: str = ""
    timeout_sec: int = 0


# '<U_IDX>' flags all the fields that cannot be simultaneously duplicated,
# '<IDX>' flags all the fields that are used in the queries
timeout_headers = {
    'company_contact': 'VARCHAR(250) NOT NULL <U_IDX>',
    'user_contact': 'VARCHAR(250) NOT NULL <U_IDX>',
    'timestamp': 'VARCHAR(50)',
    'timeout_sec': 'INT',
    }



from bot_common.utils.utils import catch_exception
from bot_common.utils.logging_conf import logger
from pydantic import BaseModel
from typing import Any
import requests
import json


class SuccessResponse(BaseModel):
    success: bool = True
    message: str = 'OK'
    response: Any = None


@catch_exception
def manage_api_request(url, req, case='', out_obj=None):
    try:
        resp = requests.post(url, json=req)
        if resp.status_code != 200:
            raise Exception(f'{case} api_request exception: error_code = {resp.status_code}')
        resp_obj = SuccessResponse.parse_obj(resp.json())
        if not bool(resp_obj.success):
            raise Exception(f'{case} exception: {str(resp_obj.message)}')
        elif out_obj:
            return out_obj.parse_obj(resp_obj.response)
        else:
            return resp_obj.response
    except requests.ConnectionError:
        raise Exception(f'{case} {url} not found')



from bot_common.utils.utils import catch_exception
from bot_common.utils.logging_conf import logger
from pydantic import BaseModel
import mysql.connector


class DbConfig(BaseModel):
    db_user: str
    db_password: str
    db_host: str
    db_port: int
    db_name: str


@catch_exception
def db_connect(conf: DbConfig):
    db = mysql.connector.connect(
        host=conf.db_host,
        port=conf.db_port,
        user=conf.db_user,
        password=conf.db_password,
        database=conf.db_name,
        charset='utf8')
    cursor = db.cursor()
    return db, cursor


@catch_exception
def create_table_if_not_exists(db, cursor, tab_name, headers):
    show_table_query = f"SHOW TABLES LIKE '{tab_name}';"
    cursor.execute(show_table_query)
    matched_tables = cursor.fetchall()

    if len(matched_tables) == 0:
        unique_index_flag = '<U_IDX>'
        index_flag = '<IDX>'
        tab_field_ls = ["pk BIGINT PRIMARY KEY AUTO_INCREMENT"]
        u_idx_field_ls = []
        idx_field_ls = []
        for h, dtype in headers.items():
            if unique_index_flag in dtype:
                u_idx_field_ls.append(h)
                dtype = dtype.replace(unique_index_flag, '').strip()
            if index_flag in dtype:
                idx_field_ls.append(f"INDEX ({h})")
                dtype = dtype.replace(index_flag, '').strip()
            tab_field_ls.append(f"{h} {dtype}")

        u_idx_fields = 'UNIQUE INDEX (' + ', '.join(u_idx_field_ls) + ')' if u_idx_field_ls else ''
        idx_fields = ', '.join(idx_field_ls)
        tab_field_ls.append(u_idx_fields)
        tab_field_ls.append(idx_fields)
        tab_field_ls = [el for el in tab_field_ls if el]
        create_table_query = f"CREATE TABLE IF NOT EXISTS {tab_name} ({', '.join(tab_field_ls)});"
        cursor.execute(create_table_query)
        db.commit()
        logger.info(f'>>> Created table: {tab_name}')
    return


@catch_exception
def insert_into_table(db, cursor, data_dict, tab_name):
    # data_dict = {k: v for k, v in data_dict.items() if v != ''}
    tab_field_ls = data_dict.keys()
    tab_value_ls = [f"'{val}'" for val in data_dict.values()]
    # compose query
    insert_into_table_query = f"INSERT INTO {tab_name} ({', '.join(tab_field_ls)}) VALUES ({', '.join(tab_value_ls)});"
    cursor.execute(insert_into_table_query)
    db.commit()
    return


@catch_exception
def update_table(db, cursor, data_dict, tab_name):
    # data_dict = {k: v for k, v in data_dict.items() if v != ''}
    # update the record
    overwrite_value_ls = [f"{h}='{val}'" for h, val in data_dict.items()]
    tab_field_ls = data_dict.keys()
    tab_value_ls = [f"'{val}'" for val in data_dict.values()]
    # compose query
    update_table_query = f"INSERT INTO {tab_name} ({', '.join(tab_field_ls)}) VALUES ({', '.join(tab_value_ls)}) ON DUPLICATE KEY UPDATE {', '.join(overwrite_value_ls)};"
    cursor.execute(update_table_query)
    db.commit()
    return



import logging

config = {
          'level': logging.DEBUG,
          'format': "%(levelname)s - %(asctime)s - %(name)s: %(message)s",
          'datefmt': '%Y-%m-%d %H:%M:%S'
         }

logger = logging.getLogger(__name__)
logging.basicConfig(**config)



from bot_common.send_message.mail_model import SendMailRequest, SenderMailConfig
from bot_common.utils import logger, catch_exception
import smtplib


class MailSender:
    def __init__(self, mail_sender: SenderMailConfig, mail_req: SendMailRequest):
        self.mail_sender = mail_sender
        self.req = mail_req

    @catch_exception
    def send(self):
        out_message = f'Subject: {self.req.mail_title}\n\n{self.req.mail_body}'
        server = smtplib.SMTP(self.mail_sender.provider_address, self.mail_sender.provider_port)
        server.starttls()
        server.login(self.mail_sender.sender_address, self.mail_sender.sender_password)
        for receiver_address in self.req.target_address_ls:
            server.sendmail(self.mail_sender.sender_address, receiver_address, out_message.encode('utf-8'))
        server.quit()



from pydantic import BaseModel


class DmMessageLightObj(BaseModel):
    session_id: str
    company_contact: str


class NewMessageDbObj(BaseModel):
    company_contact: str
    user_contact: str
    user_key: str = ''
    user_message: str = ''
    start_context: str = ''
    token: str = ''


class MessageDbObj(BaseModel):
    company_contact: str
    user_contact: str
    timestamp: str
    processed: int
    user_message: str
    start_context: str
    user_key: str


# '<U_IDX>' flags all the fields that cannot be simultaneously duplicated,
# '<IDX>' flags all the fields that are used in the queries
message_headers = {
    'company_contact': 'VARCHAR(250) NOT NULL <U_IDX>',
    'user_contact': 'VARCHAR(250) NOT NULL <U_IDX>',
    'timestamp': 'VARCHAR(50) NOT NULL <U_IDX>',
    'processed': 'TINYINT NOT NULL <IDX>',
    'user_message': 'TEXT',
    'start_context': 'TEXT',
    'user_key': 'TEXT',
    }



from bot_common.companies.company_model import config_headers, flow_headers, responses_headers, ConfigPaths
from bot_common.utils.utils import catch_exception, get_time_now
from bot_common.utils.db_utils import DbConfig, db_connect, insert_into_table, create_table_if_not_exists, update_table
from bot_common.utils.logging_conf import logger
import pandas as pd
import yaml
import json
from bot_common.lib_conf import table_name_company_configurations, flow_csv_subflow_identifier, \
    flow_csv_subflow_delimiter, flow_csv_multiple_response_ids_join, max_subflow_iterations, \
    table_prefix_conversational_flow, table_prefix_responses


# ALLOWED UP TO max_subflow_iterations LEVELS INDENTED SUBFLOWS
class FlowCsvParser:
    def __init__(self, csv_path):
        self.subflow_identifier = flow_csv_subflow_identifier.strip().lower()
        self.subflow_delimiter = flow_csv_subflow_delimiter.strip().lower()
        self.responses_separator = flow_csv_multiple_response_ids_join.strip().lower()
        self.df = pd.read_csv(csv_path, index_col=False, delimiter=';', encoding='latin-1', header=0)
        self.subflows_df = None
        self.current_row = None
        self.current_required_subflow_dict = {}
        self.current_subflow_name = ''
        self.current_in_ctx = ''
        self.other_iteration_needed = True
        self.n_iterations = 0

    @catch_exception
    def get_current_subflow_dict(self):
        try:
            self.current_required_subflow_dict = json.loads(self.current_row['out_ctx'])
        except json.decoder.JSONDecodeError:
            self.current_required_subflow_dict = {}
        except TypeError:
            self.current_required_subflow_dict = {}
        self.current_required_subflow_dict = dict(
            (k.strip().lower(), v) for k, v in self.current_required_subflow_dict.items())

    @catch_exception
    def process_current_subflow_slice(self, r):
        out_ctx = str(r['out_ctx']).strip().lower()
        r['in_ctx'] = str(r['in_ctx']).strip().lower().replace(self.current_subflow_name, self.current_in_ctx)

        if out_ctx in self.current_required_subflow_dict.keys():
            out_ls = self.current_required_subflow_dict.get(out_ctx)
            r['out_ctx'] = out_ls[0].strip().lower()
            if len(out_ls) > 1 and out_ls[1].strip():
                rid = r['response_id'] if pd.notna(r['response_id']) else ''
                r['response_id'] = rid + self.responses_separator + out_ls[1].strip()
        else:
            r['out_ctx'] = out_ctx.replace(self.current_subflow_name, self.current_in_ctx)
        return r

    @catch_exception
    def attach_subflows(self, df):
        drop_idx_all = []
        for idx, row in df.iterrows():
            self.current_row = row
            self.get_current_subflow_dict()

            if self.current_required_subflow_dict:
                idx_max = max(list(df.index))
                self.current_in_ctx = self.current_row['in_ctx'].strip().lower()
                self.current_subflow_name = self.subflow_identifier + self.subflow_delimiter + self.current_required_subflow_dict.get(self.subflow_identifier)

                subflow_df_tmp_bool = self.subflows_df['in_ctx'].apply(lambda x: str(x).strip().lower().startswith(self.current_subflow_name))
                subflow_df_tmp = self.subflows_df[subflow_df_tmp_bool]
                subflow_df_tmp_out = subflow_df_tmp.copy()
                subflow_df_tmp_out['in_ctx'] = subflow_df_tmp['in_ctx'].apply(lambda x: str(x).strip().lower().replace(self.current_subflow_name, self.current_in_ctx))
                subflow_df_tmp_out = subflow_df_tmp_out.apply(lambda r: self.process_current_subflow_slice(r), axis=1)
                drop_idx_all.append(idx)
                idx_new = list(range(idx_max + 1, idx_max + len(subflow_df_tmp_out) + 1))
                subflow_df_tmp_out.index = idx_new
                df = pd.concat([df.loc[:idx], subflow_df_tmp_out, df.loc[idx:]])

        df = df.drop(drop_idx_all, errors="ignore")
        return df.reset_index(drop=True)

    @catch_exception
    def check_other_iteration_needed(self):
        required_subflow = False
        for idx, row in self.subflows_df.iterrows():
            self.current_row = row
            self.get_current_subflow_dict()
            if self.current_required_subflow_dict:
                required_subflow = True
                break
        self.other_iteration_needed = False if not required_subflow else True
        self.n_iterations = self.n_iterations + 1 if required_subflow else self.n_iterations

    @catch_exception
    def get_subflows_df(self):
        subflow_bool = self.df['in_ctx'].apply(lambda x: str(x).strip().lower().startswith(self.subflow_identifier))
        self.subflows_df = self.df[subflow_bool]
        self.df = self.df.drop(list(self.subflows_df.index), errors="ignore").reset_index(drop=True)

        while self.other_iteration_needed:
            self.subflows_df = self.attach_subflows(self.subflows_df)
            self.check_other_iteration_needed()

            if self.n_iterations > max_subflow_iterations:
                raise Exception('Invalid fsa: infinite loop occurred due to nested subflows')

    @catch_exception
    def to_pd(self):
        non_comment_rows = self.df.iloc[:, 0].apply(lambda x: not str(x).strip().startswith('#'))
        self.df = self.df[non_comment_rows]
        self.df = self.df.dropna(axis=0, how='all')
        self.get_subflows_df()
        self.df = self.attach_subflows(self.df)
        return self.df


@catch_exception
def csv_to_pd(csv_path):
    df = pd.read_csv(csv_path, index_col=False, delimiter=';', encoding='latin-1', header=0)
    non_comment_rows = df.iloc[:, 0].apply(lambda x: not str(x).strip().startswith('#'))
    non_comment_idx = non_comment_rows[non_comment_rows].index
    df_clean = df.loc[non_comment_idx].reset_index(drop=True)
    return df_clean


@catch_exception
def parse_yaml(yaml_path):
    with open(yaml_path, 'r') as stream:
        try:
            parsed_yaml = yaml.safe_load(stream)
            return parsed_yaml
        except yaml.YAMLError as e:
            raise Exception(f"Failed to parse config file yaml {yaml_path}: {str(e)}")


class ConfigurationSetup:
    def __init__(self, company: str, db_config: DbConfig, paths: ConfigPaths):
        self.conversational_flow_table_name = table_prefix_conversational_flow + company
        self.responses_table_name = table_prefix_responses + company
        self.company_conf_table_name = table_name_company_configurations
        self.paths = paths
        self.file_path = ''
        self.tab_name = ''
        self.headers = None
        self.df = None
        self.new_pk = 1
        self.db, self.cursor = db_connect(db_config)

    @catch_exception
    def write_rows(self):
        for cnt, row in self.df.iterrows():
            row = row.dropna()
            row_data_dict = dict(zip(list(row.index), list(row.values)))
            insert_into_table(self.db, self.cursor, row_data_dict, self.tab_name)

    @catch_exception
    def drop_table_if_exists(self):
        drop_table_query = f"DROP TABLE IF EXISTS {self.tab_name};"
        self.cursor.execute(drop_table_query)
        self.db.commit()

    @catch_exception
    def setup_csv_table(self, flow=False):
        logger.info(f'>> setup table {self.tab_name}')
        self.drop_table_if_exists()
        create_table_if_not_exists(self.db, self.cursor, self.tab_name, self.headers)
        self.df = FlowCsvParser(self.file_path).to_pd() if flow else csv_to_pd(self.file_path)
        self.df = self.df.applymap(lambda x: x.replace("\\", "\\\\").replace("'", "\\'") if isinstance(x, str) else x)
        self.write_rows()

    @catch_exception
    def get_new_pk(self):
        get_pks_query = f"SELECT pk FROM {self.tab_name}"
        self.cursor.execute(get_pks_query)
        pks = self.cursor.fetchall()
        self.new_pk = max([el[0] for el in pks]) + 1 if pks else 1

    @catch_exception
    def setup_yaml_config(self):
        logger.info(f'>> setup table {self.tab_name}')
        create_table_if_not_exists(self.db, self.cursor, self.tab_name, self.headers)
        data_dict_var = parse_yaml(self.file_path).get('Variables')
        data_dict = {}
        field_other = {}
        for k, v in data_dict_var.items():
            v = json.dumps(v) if (isinstance(v, list) or isinstance(v, dict)) else v
            v = v.replace("\\", "\\\\").replace("'", "\\'") if isinstance(v, str) else v
            if k in self.headers.keys():
                data_dict[k] = v
            else:
                field_other[k] = v
        self.get_new_pk()
        data_dict['pk'] = self.new_pk
        data_dict['is_active'] = 1
        data_dict['is_active_switch_tmp'] = get_time_now()
        data_dict['other'] = json.dumps(field_other)
        update_table(self.db, self.cursor, data_dict, self.tab_name)

    # --- MAIN ---

    @catch_exception
    def setup(self):
        self.file_path = self.paths.flow_path
        self.tab_name = self.conversational_flow_table_name
        self.headers = flow_headers
        self.setup_csv_table(flow=True)

        self.file_path = self.paths.responses_path
        self.tab_name = self.responses_table_name
        self.headers = responses_headers
        self.setup_csv_table(flow=False)

        self.file_path = self.paths.config_file_path
        self.tab_name = self.company_conf_table_name
        self.headers = config_headers
        self.setup_yaml_config()
        self.db.close()



table_name_company_configurations = 'configurations'
table_name_exceptions = 'sessions_exceptions'
table_name_messages = 'messages'
table_name_timeout = 'timeout'

table_prefix_conversational_flow = 'conversational_flow_'
table_prefix_responses = 'responses_'
table_prefix_sessions = 'sessions_'
table_prefix_logs = 'logs_'
table_prefix_logs_end_conversation = 'logs_endconv_'

flow_csv_multiple_response_ids_join = '&'
flow_csv_subflow_identifier = 'subflow'
flow_csv_subflow_delimiter = '|'

void_response = {'RV': '/'}
max_subflow_iterations = 5



from bot_common.companies.company import Company, CompanyManager, CompanyTables
from bot_common.companies.company_setup import ConfigurationSetup
from bot_common.companies.company_model import ConfigPaths



from pydantic import BaseModel


class ExceptionObj(BaseModel):
    timestamp: str = ''
    state: str = ''
    error_message: str = ''


# '<U_IDX>' flags all the fields that cannot be simultaneously duplicated,
# '<IDX>' flags all the fields that are used in the queries
exception_headers = {'timestamp': 'DATETIME(6) NOT NULL <IDX>',
                     'state': 'TEXT',
                     'error_message': 'TEXT',
                     }



from pydantic import BaseModel


class LogObject(BaseModel):
    id_session: str
    company: str
    company_contact: str
    user_key: str = ""
    start_context: str = ""
    current_context: str = ""
    conversation_platform: str = ""
    start_conversation_timestamp: str = ""
    timestamp: str = ""
    conv_duration_sec: int = 0
    detected_intent: str = ""
    intent_confidence: float = 0
    conv_step_num: int = 0
    closed_formality: bool = False
    unclosed_success: bool = False
    platform_exception: bool = False
    solicit: bool = False
    fallback: bool = False
    handover: bool = False
    handover_incomprehension: bool = False
    closed: bool = False
    hangup: bool = False
    redirect: bool = False
    expired: bool = False
    state: str = ""
    message_id: str = ""
    bot_message: str = ""
    current_user_utterance: str = ""
    log_transcript: str = ""
    other_logs: dict = {}
    extracted_data: dict = {}


# '<U_IDX>' flags all the fields that cannot be simultaneously duplicated,
# '<IDX>' flags all the fields that are used in the queries
log_headers = {'timestamp': 'DATETIME(6) NOT NULL <U_IDX>',
               'id_session': 'VARCHAR(250) NOT NULL <U_IDX>',
               'company': 'VARCHAR(250) NOT NULL <IDX>',
               'company_contact': 'VARCHAR(250) NOT NULL <IDX>',
               'user_key': 'VARCHAR(250) <IDX>',
               'start_context': 'VARCHAR(250) <IDX>',
               'current_context': 'TEXT',
               'conversation_platform': 'VARCHAR(50)',
               'start_conversation_timestamp': 'DATETIME(6) NOT NULL <IDX>',
               'conv_duration_sec': 'INT',
               'detected_intent': 'TEXT',
               'intent_confidence': 'DOUBLE(3,2)',
               'conv_step_num': 'TINYINT <IDX>',
               'closed_formality': 'TINYINT <IDX>',
               'unclosed_success': 'TINYINT <IDX>',
               'platform_exception': 'TINYINT <IDX>',
               'solicit': 'TINYINT <IDX>',
               'fallback': 'TINYINT <IDX>',
               'handover': 'TINYINT <IDX>',
               'handover_incomprehension': 'TINYINT <IDX>',
               'closed': 'TINYINT <IDX>',
               'hangup': 'TINYINT <IDX>',
               'redirect': 'TINYINT <IDX>',
               'expired': 'TINYINT <IDX>',
               'state': 'VARCHAR(250)',
               'message_id': 'TEXT',
               'bot_message': 'TEXT',
               'current_user_utterance': 'TEXT',
               'log_transcript': 'TEXT',
               'other_logs': 'TEXT',
               'extracted_data': 'TEXT',
               }



from bot_common.sessions.session_model import SessionObject, session_headers as headers
from bot_common.utils.db_utils import DbConfig, create_table_if_not_exists, db_connect, update_table
from bot_common.utils.utils import catch_exception, get_time_now, clean_string
from bot_common.utils.logging_conf import logger
from bot_common.lib_conf import table_prefix_sessions
from typing import List, Union
import json

# - In "cache" we store some cached_variables for the backend logic
#   and the printable_variables for the responses.
# - In "other_logs" we store some indicators about the performed
#   conversational flow path.
# - In "extracted_data" we store the main features extracted
#   during the conversation.

int_fields = ['timeout_sec', 'bot_message_contains_buttons', 'unclosed_success',
              'solicited_times', 'is_closed', 'platform_exception']
dump_fields = ['cache', 'entities', 'other_logs', 'extracted_data']
str_special_char_fields = ['bot_message']
clean_str_fields = []
float_fields = []


class Session(SessionObject):
    @catch_exception
    def extract_data(self):
        data = self.__dict__.copy()
        data['timestamp'] = get_time_now()
        for field in dump_fields:
            data[field] = clean_string(json.dumps(data[field])).replace('\\', '\\\\')
        for field in str_special_char_fields:
            data[field] = data[field].replace("\\", "\\\\").replace("'", "\\'")
        for field in int_fields:
            data[field] = int(data[field])
        for field in clean_str_fields:
            data[field] = clean_string(data[field])
        for field in float_fields:
            data[field] = round(float(data[field]), 2)
        return data

    @catch_exception
    def set(self, db_config: DbConfig):
        tab_name = table_prefix_sessions + self.company
        session_db, session_cursor = db_connect(db_config)
        create_table_if_not_exists(session_db, session_cursor, tab_name, headers)
        data_dict = self.extract_data()
        update_table(session_db, session_cursor, data_dict, tab_name)
        logger.info(f'set session {self.id_session} success')
        session_db.close()

    @catch_exception
    def delete(self, db_config: DbConfig):
        tab_name = table_prefix_sessions + self.company
        session_db, session_cursor = db_connect(db_config)
        create_table_if_not_exists(session_db, session_cursor, tab_name, headers)
        delete_session_query = f"DELETE FROM {tab_name} WHERE id_session = '{self.id_session}';"
        session_cursor.execute(delete_session_query)
        session_db.commit()
        logger.info(f'delete session {self.id_session} success')
        session_db.close()

    @catch_exception
    def close(self, db_config: DbConfig):
        tab_name = table_prefix_sessions + self.company
        session_db, session_cursor = db_connect(db_config)
        create_table_if_not_exists(session_db, session_cursor, tab_name, headers)
        data_dict = self.extract_data()
        data_dict['is_closed'] = 1
        update_table(session_db, session_cursor, data_dict, tab_name)
        logger.info(f'close session {self.id_session} success')
        session_db.close()


class SessionGet:
    @catch_exception
    def __init__(self, company: str, db_config: DbConfig):
        self.company = company
        self.tab_name = table_prefix_sessions + company
        self.session_db, self.session_cursor = db_connect(db_config)
        create_table_if_not_exists(self.session_db, self.session_cursor, self.tab_name, headers)

    @catch_exception
    def single(self, session_id: str) -> Union[Session, None]:
        get_session_query = f"SELECT * FROM {self.tab_name} WHERE id_session = '{session_id}';"
        self.session_cursor.execute(get_session_query)
        myresult = self.session_cursor.fetchall()
        self.session_db.close()
        if not myresult:
            logger.info(f"session ({session_id}) not found")
            return None
        # myresult[0][1:] because we need to remove the pk auto_increment element
        out_dict = dict(zip(list(headers.keys()), myresult[0][1:]))
        for field in dump_fields:
            out_dict[field] = json.loads(out_dict[field])
        return Session.parse_obj(out_dict)

    @catch_exception
    def existing_all(self) -> List[Session]:
        get_existing_sessions_query = f"SELECT * FROM {self.tab_name};"
        self.session_cursor.execute(get_existing_sessions_query)
        myresult = self.session_cursor.fetchall()
        self.session_db.close()
        parsed_out = []
        for res in myresult:
            # res[1:] is needed to remove the pk auto_increment key
            parsed_dict = dict(zip(list(headers.keys()), res[1:]))
            for field in dump_fields:
                parsed_dict[field] = json.loads(parsed_dict[field])
            parsed_obj = Session.parse_obj(parsed_dict)
            parsed_out.append(parsed_obj)
        return parsed_out

    @catch_exception
    def delete_table(self):
        delete_sessions_table_query = f"DROP TABLE {self.tab_name};"
        self.session_cursor.execute(delete_sessions_table_query)
        self.session_db.commit()
        logger.info(f'drop table {self.tab_name} success')
        self.session_db.close()
